package main

import (
    "os"
    "bytes"
    "errors"
    "strings"
    _ "embed"
    "os/exec"
    "encoding/json"
    "nexus-common/vm"
    "nexus-common/params"
    e "nexus-client/exec"
    u "nexus-client/utils"
    g "nexus-client/globals"
    // "nexus-client/cmdVersion"
    "nexus-client/defaults"
    "nexus-client/cmdLogin"
    "fyne.io/fyne/v2"
    "fyne.io/fyne/v2/app"
    "fyne.io/fyne/v2/theme"
    "fyne.io/fyne/v2/widget"
    "fyne.io/fyne/v2/dialog"
    "fyne.io/fyne/v2/container"
    "github.com/go-resty/resty/v2"
    "github.com/go-playground/validator/v10"
)

const (
    windowTitle = "nexus-exam"
)

var (
    //go:embed nexus_exam_user.val
    nexus_exam_user string

    //go:embed nexus_exam_pwd.val
    nexus_exam_pwd string

    certPath string
    exitFn func()
)

func exit(code int) {
    if exitFn != nil {
        exitFn()
    }
    os.Exit(code)
}

func errorDialog(parent fyne.Window, msg string) {
    dialog.NewInformation("Connection error", msg, parent).Show()
}

func attachVM(parent fyne.Window, token, hostname, cert, pwd string) (int, error) {
    client := g.GetInstance().Client
    host := g.GetInstance().Host

    client.SetHeader("Content-Type", "application/json")
    client.SetHeader("Authorization", "Bearer "+token)

    p := &params.VMAttachCreds{ Pwd: pwd }
    resp, err := client.R().SetBody(p).Post(host+"/vms/spicecreds")
    if err != nil {
        errorDialog(parent, err.Error())
    }

    if resp.IsSuccess() {
        var creds vm.VMSpiceCredentialsSerialized
        if err := json.Unmarshal(resp.Body(), &creds); err != nil {
            errorDialog(parent, err.Error())
        }
        if err := validator.New(validator.WithRequiredStructEnabled()).Struct(creds); err != nil {
            errorDialog(parent, err.Error())
        }
        go func(creds vm.VMSpiceCredentialsSerialized) {
            _, err := e.RunRemoteViewer(hostname, cert, creds.Name, creds.SpicePort, creds.SpicePwd, false)
            if err != nil {
                errorDialog(parent, err.Error())
            }
        } (creds)
    } else {
        type msg struct {
            Message string  `json:"message"`
        }
        var m msg
        if err := json.Unmarshal(resp.Body(), &m); err != nil {
            errorDialog(parent, err.Error())
        }
        errorDialog(parent, m.Message)
        // errorDialog(parent, "Error: "+resp.Status()+": "+resp.String())
    }
    return 0, nil
}

func abortWindow(labelMsg string) {
	a := app.New()
    a.Settings().SetTheme(theme.LightTheme())
	win := a.NewWindow(windowTitle)
    win.SetOnClosed(func() {
        win.Close()
		exit(1)
    })
	label := widget.NewLabel("ERROR: "+labelMsg)
	button := widget.NewButton("OK", func() {
        win.Close()
		exit(1)
	})
    content := container.NewPadded(container.NewVBox(label, button))
	win.SetContent(content)
    win.ShowAndRun()
}

func hypervisorCheck() {
    cmd := exec.Command("systemd-detect-virt")
    var out bytes.Buffer
    cmd.Stdout = &out
    cmd.Run()
    // Returns "none\n" if running on a physical system (no VM/container)
    if out.String() != "none\n" {
        abortWindow("nexus-exam only works on a physical system!")
    }
}

func run() int {
    hypervisorCheck()

    if err := e.CheckRemoteViewer(); err != nil {
        abortWindow(err.Error())
    }

    serverEnvVar, found := os.LookupEnv(g.ENV_NEXUS_SERVER)
    if !found {
        serverEnvVar = defaults.NexusServer
    }

    var err error  // necessary for certPath below to be ref as the same variable

    certPath, found = os.LookupEnv(g.ENV_NEXUS_CERT)
    if !found {
        certPath, err = defaults.CreateCert()
        if err != nil {
            abortWindow("Failed creating certificate from embedded certificate!")
        }
        // No embedded certificate, exit with a information message.
        if certPath == "" {
            abortWindow("environment variable \""+g.ENV_NEXUS_CERT+"\" is not set!")
        }
        exitFn = func() { os.Remove(certPath) }

        // This thread acts as a signal handler for SIGINT or SIGTERM.
        // When one of these signals is received, the temporary certificate file is deleted.
        // Without this "handler", the temporary certificate file wouldn't be deleted.
        go func() {
            u.WaitForSignals()
            exit(1)
        }()
    }

    parts := strings.Split(serverEnvVar, ":")
    hostname := parts[0]

    client := resty.New()
    client.SetRootCertificate(certPath)
    host := "https://"+serverEnvVar

    g.Init(hostname, host, certPath, client)

    // Checks the client version is compatible with the server's API.
    // if !cmdVersion.CheckServerCompatibility("nexus-exam") {
    //     abortWindow("client version is incompatible with server!")
    // }

    // Logins and obtains a JWT token.
    token, err := cmdLogin.GetToken(nexus_exam_user, nexus_exam_pwd)
    if err != nil {
        abortWindow(err.Error())
    }

    app := app.New()
    app.Settings().SetTheme(theme.LightTheme())
    win := app.NewWindow(windowTitle)

    pwdEntry := widget.NewEntry()
    pwdEntry.Password = true

    pwdEntry.Validator = func(val string) error {
        if len(val) < 8 {
            return errors.New("Please enter at least 8 characters")
        }
        return nil
    }   

    label := widget.NewLabel("Enter password to connect to your VM:")

    form := &widget.Form{
        Items: []*widget.FormItem {
            {Text: "Password", Widget: pwdEntry},
        },
        OnSubmit: func() {
            attachVM(win, token, hostname, certPath, pwdEntry.Text)
        },
        SubmitText: "Connect",
    }

    win.SetContent(container.NewPadded(container.NewVBox(label, form)))
    win.Resize(fyne.NewSize(600,200))
    win.ShowAndRun()
    return 0
}

func main() {
    exit(run())
}
